import * as apis from '@vnmfjs/vnmf-h5'
import * as babel from 'babel-core'
import * as t from 'babel-types'

import plugin from '../src'

const pluginOptions = [
  plugin,
  {
    apis,
    packageName: '@vnmfjs/vnmf-h5'
  }
]

const getNamedImports = (importSpecifiers: (t.ImportSpecifier | t.ImportDefaultSpecifier | t.ImportNamespaceSpecifier)[]) => {
  return importSpecifiers.reduce((prev, curr: t.ImportSpecifier) => {
    if (t.isImportSpecifier(curr)) {
      prev.add(curr.imported.name)
    }
    return prev
  }, new Set())
}

it('should work!', function () {
  const code = `
    import Vnmf, { setStorage, initPxTransform, param } from '@vnmfjs/vnmf-h5';
    initPxTransform(param)
    Vnmf.initPxTransform()
    Vnmf.initPxTransform()
    Vnmf['getStorage']()
    setStorage()
    export { Vnmf }
  `
  const result = babel.transform(code, { plugins: [pluginOptions] })
  expect(result.code).toMatchSnapshot()
})

it('should leave other apis untouched', function () {
  const code = `
    import Vnmf from '@vnmfjs/vnmf-h5'
    Vnmf.noop
  `
  const result = babel.transform(code, { plugins: [pluginOptions] })
  expect(result.code).toMatchSnapshot()

  const ast = result.ast as t.File
  const body = ast.program.body as [t.ImportDeclaration, t.ExpressionStatement]
  expect(t.isImportDeclaration(body[0])).toBeTruthy()
  expect(t.isExpressionStatement(body[1])).toBeTruthy()
  const defaultImport = body[0].specifiers.find(v => t.isImportDefaultSpecifier(v))
  expect(defaultImport).toBeTruthy()

  const vnmfName = defaultImport.local.name
  const namedImports = getNamedImports(body[0].specifiers)
  expect(namedImports).toEqual(new Set())
  expect(t.isMemberExpression(body[1].expression)).toBeTruthy()
  expect((body[1].expression as t.MemberExpression)).toMatchObject(t.memberExpression(
    t.identifier(vnmfName),
    t.identifier('noop')
  ))
})

it('should move static apis under "Vnmf"', function () {
  const code = `
    import { noop } from '@vnmfjs/vnmf-h5';
    noop;
    noop();
  `

  const result = babel.transform(code, { plugins: [pluginOptions] })
  expect(result.code).toMatchSnapshot()

  const ast = result.ast as t.File
  const body = ast.program.body as [t.ImportDeclaration, t.ExpressionStatement]
  expect(t.isImportDeclaration(body[0])).toBeTruthy()
  expect(t.isExpressionStatement(body[1])).toBeTruthy()
  const defaultImport = body[0].specifiers.find(v => t.isImportDefaultSpecifier(v))
  expect(defaultImport).toBeTruthy()

  const vnmfName = defaultImport!.local.name
  let memberExpression = body[1].expression
  if (t.isCallExpression(body[1])) {
    memberExpression = (body[1].expression as t.CallExpression).callee
  }
  expect(memberExpression).toMatchObject(t.memberExpression(
    t.identifier(vnmfName),
    t.identifier('noop')
  ))
})

it('should not import vnmf duplicatly', function () {
  const code = `
    import { Component } from "@vnmfjs/vnmf-h5";
    import Vnmf from '@vnmfjs/vnmf-h5';
    Component
    Vnmf.createAnimation()
    Vnmf.initPxTransform()
  `

  const result = babel.transform(code, { plugins: [pluginOptions] })
  expect(result.code).toMatchSnapshot()

  const ast = result.ast as t.File
  const body = ast.program.body as [t.ImportDeclaration, t.ExpressionStatement, t.ExpressionStatement]
  expect(t.isImportDeclaration(body[0])).toBeTruthy()
  expect(t.isExpressionStatement(body[1])).toBeTruthy()
  expect(t.isExpressionStatement(body[2])).toBeTruthy()
})

it('should not go wrong when using an api twice', function () {
  const code = `
    import Vnmf from '@vnmfjs/vnmf-h5';
    const animation = Vnmf.createAnimation({
      duration: dura * 1000,
      timingFunction: 'linear'
    })
    const resetAnimation = Vnmf.createAnimation({
      duration: 0,
      timingFunction: 'linear'
    })
  `
  expect(() => {
    const result = babel.transform(code, { plugins: [pluginOptions] })
    expect(result.code).toMatchSnapshot()
  }).not.toThrowError()
})

it('should preserve default imports', function () {
  const code = `
    import Vnmf from '@vnmfjs/vnmf-h5'
    console.log(Vnmf)
  `
  const result = babel.transform(code, { plugins: [pluginOptions] })
  expect(result.code).toMatchSnapshot()
})

it('should preserve assignments in lefthands', function () {
  const code = `
    import Vnmf from '@vnmfjs/vnmf-h5'
    let animation
    animation = Vnmf.createAnimation({
      transformOrigin: "50% 50%",
      duration: 1000,
      timingFunction: "ease",
      delay: 0
    });
    Vnmf.request()
    Vnmf.request = ''
    Vnmf['request'] = ''
  `
  const result = babel.transform(code, { plugins: [pluginOptions] })
  expect(result.code).toMatchSnapshot()
})

it('should support rename of imported names', function () {
  const code = `
  // import { inject as mobxInject, observer as mobxObserver } from '@vnmfjs/mobx'
  import { Component as VnmfComponent } from "@vnmfjs/vnmf-h5";
  export class Connected extends VnmfComponent {}
  `
  const result = babel.transform(code, { plugins: [pluginOptions] })
  expect(result.code).toMatchSnapshot()
})
